/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#include "gdamrawfile.h"

typedef struct _GdamSubRawFile GdamSubRawFile;
struct _GdamSubRawFile {
	GdamSource		source;
	FILE*			fp;
	int			sample_size;
	gint64			bytes_remaining;
	void		      (*converter)(GdamSample* sample,
	                                   const void* input,
					   int         sample_len);
	void                  (*close_func)(FILE*);
};

typedef void	         (*GdamSampleConverter)       (GdamSample* sample,
	                                               const void* input,
					               int         sample_len);
static GdamSampleConverter gdam_raw_file_get_converter(int         flags,
                                                       int*        samp_len);
static GdamSubRawFile*     gdam_sub_raw_file_new      (FILE*       fp,
	                                               int	   sample_size,
	                                               GdamSampleConverter,
					               void (*closer)(FILE*));
static int                 gdam_sub_raw_file_read     (GdamSource* source,
                                                       GdamSample* buf,
						       int         n);
static void                gdam_sub_raw_file_miss_data(GdamSource* source,
                                                       int         n);
static void                gdam_sub_raw_file_destroy  (GtkObject*  object);

#define GDAM_TYPE_SUB_RAW_FILE              (gdam_sub_raw_file_get_type ())
#define GDAM_SUB_RAW_FILE(obj)              (GTK_CHECK_CAST ((obj), GDAM_TYPE_SUB_RAW_FILE, GdamSubRawFile))
static GtkType gdam_sub_raw_file_get_type();

static GtkObjectClass* resample_class = NULL;
static GtkObjectClass* source_class = NULL;

/*
 *                        __ _ _
 *  _ __ __ ___      __  / _(_) | ___
 * | '__/ _` \ \ /\ / / | |_| | |/ _ \
 * | | | (_| |\ V  V /  |  _| | |  __/
 * |_|  \__,_| \_/\_/   |_| |_|_|\___|
 * 
 */

GtkType gdam_raw_file_get_type()
{
	static GtkType source_type = 0;
	if (!source_type) {
		static const GtkTypeInfo source_info = {
			"GdamRawFile",
			sizeof(GdamRawFile),
			sizeof(GdamResampleClass),
			(GtkClassInitFunc) NULL,
			(GtkObjectInitFunc) NULL,
			/* reserved_1*/ NULL,
			/* reserved_2*/ NULL,
			(GtkClassInitFunc) NULL
		};
		resample_class = gtk_type_class(GDAM_TYPE_RESAMPLE);
		source_type = gtk_type_unique(GDAM_TYPE_RESAMPLE,
		                              &source_info);
	}
	return source_type;
}


void gdam_raw_file_configure(GdamRawFile*  raw_file,
                             FILE*         fp,
			     void   (*close_func)(FILE*),
			     int           input_hertz,
			     int           sample_type)
{
	GdamResample* resample = GDAM_RESAMPLE(raw_file);
	gdouble rate;
	GdamFilter* filter = GDAM_FILTER(raw_file);
	GdamSampleConverter converter;
	int length;
	GdamSource* subsource;
	GdamSubRawFile* sub_raw_file;
	g_return_if_fail(resample != NULL && filter != NULL);
	g_return_if_fail(fp != NULL);

	converter = gdam_raw_file_get_converter(sample_type, &length);
	if (converter == NULL) {
		gdam_debug("no converter for input sample type found");
		return;
	}
	sub_raw_file = gdam_sub_raw_file_new(fp, length, converter, close_func);
	sub_raw_file->bytes_remaining = raw_file->bytes_remaining;
	subsource = GDAM_SOURCE(sub_raw_file);
	g_return_if_fail(subsource);
	gdam_filter_set_source(filter, subsource);
	rate = (gdouble)input_hertz / 44100.0 * (65536.0*65536.0);
	resample->rate = rate;
}

/*
 *            _                             __ _ _
 *  ___ _   _| |__    _ __ __ ___      __  / _(_) | ___
 * / __| | | | '_ \  | '__/ _` \ \ /\ / / | |_| | |/ _ \
 * \__ \ |_| | |_) | | | | (_| |\ V  V /  |  _| | |  __/
 * |___/\__,_|_.__/  |_|  \__,_| \_/\_/   |_| |_|_|\___|
 * 
 */
 
static void gdam_sub_raw_file_class_init(GdamSourceClass* source_class)
{
	source_class->read = gdam_sub_raw_file_read;
	source_class->miss_data = gdam_sub_raw_file_miss_data;
	((GtkObjectClass*)source_class)->destroy = gdam_sub_raw_file_destroy;
}

#ifdef gdam_sample_copy
#undef gdam_samply_copy
#endif

static void gdam_sub_raw_file_init(GdamSubRawFile* sub_raw_file)
{
	sub_raw_file->sample_size = 4;
	sub_raw_file->bytes_remaining = -1;
	sub_raw_file->converter = (GdamSampleConverter) gdam_sample_copy;
}

static int gdam_sub_raw_file_read(GdamSource* source, GdamSample* buf, int n)
{
	int i;
	GdamSubRawFile* sub_raw_file = GDAM_SUB_RAW_FILE(source);
	void* input = alloca(n * sub_raw_file->sample_size);
	if (sub_raw_file->bytes_remaining >= 0) {
		gint64 samples_remaining = sub_raw_file->bytes_remaining;
		samples_remaining /= sub_raw_file->sample_size;
		if (samples_remaining < n)
			n = samples_remaining;
	}
	if (n <= 0)
		return n;
	if ((i = fread(input,
	               sub_raw_file->sample_size,
		       n,
		       sub_raw_file->fp)) <= 0)
		return 0;
	if (sub_raw_file->bytes_remaining > 0)
		sub_raw_file->bytes_remaining -= i * sub_raw_file->sample_size;
	if (sub_raw_file->converter)
		(*sub_raw_file->converter)(buf, input, i);
	return i;
}
static void gdam_sub_raw_file_miss_data(GdamSource* source, int n)
{
	GdamSubRawFile* sub = GDAM_SUB_RAW_FILE(source);
	if (fseek(sub->fp, n * sub->sample_size, SEEK_CUR) < 0) {
		void* buf = alloca(sub->sample_size * n);
		fread(buf, sub->sample_size, n, sub->fp);
	}
}

static GtkType gdam_sub_raw_file_get_type()
{
	static GtkType sub_raw_type = 0;
	if (!sub_raw_type) {
		static const GtkTypeInfo sub_raw_info = {
			"GdamSubRawFile",
			sizeof(GdamSubRawFile),
			sizeof(GdamResampleClass),
			(GtkClassInitFunc) gdam_sub_raw_file_class_init,
			(GtkObjectInitFunc) gdam_sub_raw_file_init,
			/* reserved_1*/ NULL,
			/* reserved_2*/ NULL,
			(GtkClassInitFunc) NULL
		};
		source_class = gtk_type_class(GDAM_TYPE_SOURCE);
		sub_raw_type = gtk_type_unique(GDAM_TYPE_SOURCE,&sub_raw_info);
	}
	return sub_raw_type;
}
static void gdam_sub_raw_file_destroy(GtkObject*  object)
{
	GdamSubRawFile* sub_raw_file = (GdamSubRawFile*)object;
	if (sub_raw_file->fp) {
		(*sub_raw_file->close_func)(sub_raw_file->fp);
		sub_raw_file->fp = NULL;
	}
	(*resample_class->destroy)(object);
}

static GdamSubRawFile* gdam_sub_raw_file_new(FILE*               fp,
	                                     int	         sample_size,
	                                     GdamSampleConverter converter,
					     void    (*closer)(FILE*))
{
	GdamSubRawFile* file;
	file = GDAM_SUB_RAW_FILE(gtk_type_new(gdam_sub_raw_file_get_type()));
	file->fp = fp;
	file->sample_size = sample_size;
	file->converter = converter;
	file->close_func = closer;
	return file;
}

/*=====================================================================*/
static void convert_from_8s_stereo(GdamSample* _out, const void* in, int len)
{
	const gint8* at = (const gint8*)in;
	GdamSampleValue* out = (GdamSampleValue*)_out;
	len += len;
	while (len--) {
		guint8 i = (guint8)(*at);
		*out = ((GdamSampleValue)i << (GDAM_SAMPLE_LOG - 7));
		out++; at++;
	}
}
static void convert_from_8u_stereo(GdamSample* _out, const void* in, int len)
{
	const gint8* at = (const gint8*)in;
	GdamSampleValue* out = (GdamSampleValue*)_out;
	len += len;
	while (len--) {
		guint8 i = (guint8)(*at ^ 0x80);
		*out = ((GdamSampleValue)i << (GDAM_SAMPLE_LOG - 7));
		out++; at++;
	}
}
static void convert_from_8s_mono(GdamSample* _out, const void* in, int len)
{
	const gint8* at = (const gint8*)in;
	GdamSampleValue* out = (GdamSampleValue*)_out;;
	while (len--) {
		guint8 i = (guint8)(*at);
		out[0] = out[1] = ((GdamSampleValue)i << (GDAM_SAMPLE_LOG - 7));
		out += 2; at++;
	}
}
static void convert_from_8u_mono(GdamSample* _out, const void* in, int len)
{
	const gint8* at = (const gint8*)in;
	GdamSampleValue* out = (GdamSampleValue*)_out;
	while (len--) {
		guint8 i = (guint8)(*at ^ 0x80);
		out[0] = out[1] = ((GdamSampleValue)i) << (GDAM_SAMPLE_LOG - 7);
		out += 2; at++;
	}
}

static void convert_from_16u_stereo(GdamSample* _out, const void* in, int len)
{
	const gint16* at = (const gint16*)in;
	GdamSampleValue* out = (GdamSampleValue*)_out;
	len += len;
	while (len--) {
		gint16 val = *at ^ 0x8000;
		*out = ((GdamSampleValue)val) << (GDAM_SAMPLE_LOG - 15);
		out++; at++;
	}
}
static void convert_from_16s_mono(GdamSample* _out, const void* in, int len)
{
	const gint16* at = (const gint16*)in;
	GdamSampleValue* out = (GdamSampleValue*)_out;
	while (len--) {
		out[0] = out[1] = ((GdamSampleValue)*at) << (GDAM_SAMPLE_LOG-15);
		out += 2; at++;
	}
}
static void convert_from_16u_mono(GdamSample* _out, const void* in, int len)
{
	const gint16* at = (const gint16*)in;
	GdamSampleValue* out = (GdamSampleValue*)_out;
	while (len--) {
		gint16 val = *at ^ 0x8000;
		out[0] = out[1] = ((GdamSampleValue)val) << (GDAM_SAMPLE_LOG-15);
		out += 2; at++;
	}
}
static void convert_from_16s_stereo(GdamSample* _out, const void* in, int len)
{
	const gint16* at = (const gint16*)in;
	GdamSampleValue* out = (GdamSampleValue*)_out;;
	len += len;
	while (len--) {
		out[0] = ((GdamSampleValue)at[0] << (GDAM_SAMPLE_LOG-15));
		out++;
		at++;
	}
}

static GdamSampleConverter converter_table[] = {
	convert_from_8u_mono,
	convert_from_8s_mono,
	convert_from_8u_stereo,
	convert_from_8s_stereo,
	convert_from_16u_mono,
	convert_from_16s_mono,
	convert_from_16u_stereo,
	convert_from_16s_stereo,
};

static GdamSampleConverter gdam_raw_file_get_converter(int  flags, int* rv_len)
{
	int index = 0;
	int len = 1;
	if ((flags & GDAM_RAW_FILE_SIGN_MASK) == GDAM_RAW_FILE_SIGNED)
		index += 1;
	else if ((flags & GDAM_RAW_FILE_SIGN_MASK) == GDAM_RAW_FILE_UNSIGNED)
		;
	else {
		gdam_debug("sample style was neither signed nor unsigned");
		return NULL;
	}
	if ((flags & GDAM_RAW_FILE_SIZE_MASK) == GDAM_RAW_FILE_16BIT) {
		index += 4;
		len += len;
	} else if ((flags & GDAM_RAW_FILE_SIZE_MASK) == GDAM_RAW_FILE_8BIT)
		;
	else {
		gdam_debug("sample size was neither 8 nor 16");
		return NULL;
	}
	if ((flags & GDAM_RAW_FILE_CHANNEL_MASK) == GDAM_RAW_FILE_MONO)
		;
	else if ((flags&GDAM_RAW_FILE_CHANNEL_MASK) == GDAM_RAW_FILE_STEREO) {
		index += 2;
		len += len;
	} else {
		gdam_debug("sample style was neither signed nor unsigned");
		return NULL;
	}
	*rv_len = len;
	gdam_debug("raw_file: using converter at index %d", index);
	return converter_table[index];
}

void gdam_raw_file_set_bytes_remaining(GdamRawFile*    raw_file,
                                       gint64          bytes_remaining)
{
	GdamSource* subsource = GDAM_FILTER(raw_file)->filtered;
	raw_file->bytes_remaining = bytes_remaining;
	if (subsource != NULL)
		GDAM_SUB_RAW_FILE(subsource)->bytes_remaining = bytes_remaining;
}

