/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#include "gdamwavetable.h"
#include <gdam/gdam-protocol.h>

static int         gdam_wavetable_read             (GdamSource*   input,
	                                            GdamSample*   samples,
						    int           max_read);
static void        gdam_wavetable_miss_data        (GdamSource*   input,
	                                            int           samples_lost);
static GdamTime    gdam_wavetable_get_position     (GdamSource*   source);
static gboolean    gdam_wavetable_seek             (GdamSource*   source,
                                                    GdamTime      time,
						    int           whence);

static void        gdam_wavetable_destroy          (GtkObject*    object);

static GtkObjectClass* parent_class = NULL;
static GdamSourceClass* parent_source_class = NULL;

static void gdam_wavetable_arg_set_hertz(GtkObject* object, GdamArg* arg)
{
	GdamWavetable* wavetable = GDAM_WAVETABLE(object);
	int buffer_len;
	double hertz = GDAM_ARG_DOUBLE(arg);
	
	if (wavetable->cache == NULL)
		return;

	buffer_len = wavetable->cache->buffer_length;
	hertz *= (1<<16);
	hertz *= (1<<16);
	hertz /= 44100.0;
	hertz *= buffer_len;
	wavetable->velocity_32 = (gint64)(hertz);
	gdam_debug("wavetable_set_hertz: velocity=%lld",
	           wavetable->velocity_32);
}
static void gdam_wavetable_arg_get_hertz(GtkObject* object, GdamArg* arg)
{
	GdamWavetable* wavetable = GDAM_WAVETABLE(object);
	int buffer_len = wavetable->cache->buffer_length;
	double hertz = wavetable->velocity_32;
	hertz /= (1<<16);
	hertz /= (1<<16);
	hertz *= 44100.0;
	hertz /= buffer_len;
	GDAM_ARG_DOUBLE(arg) = hertz;
}
static void gdam_wavetable_arg_set_velocity(GtkObject* object, GdamArg* arg)
{
	GdamWavetable* wavetable = GDAM_WAVETABLE(object);
	double velocity = GDAM_ARG_DOUBLE(arg);
	velocity *= (1<<16);
	velocity *= (1<<16);
	wavetable->velocity_32 = (gint64)velocity;
}
static void gdam_wavetable_arg_get_velocity(GtkObject* object, GdamArg* arg)
{
	GdamWavetable* wavetable = GDAM_WAVETABLE(object);
	double velocity = wavetable->velocity_32;
	velocity /= (1<<16);
	velocity /= (1<<16);
	GDAM_ARG_DOUBLE(arg) = velocity;
}

static void gdam_wavetable_arg_set_buffer_id(GtkObject*        object,
                                             GdamArg*          arg)
{
	GdamCacheBuffer* cache_buf;
	GdamWavetable* wavetable = GDAM_WAVETABLE(object);
	int buffer_id = GDAM_ARG_INT(arg);
	cache_buf = gdam_cache_buffer_get_by_id(buffer_id);
	if (cache_buf == NULL) {
		gdam_debug("no buffer id %d", buffer_id);
		return;
	}
	if (cache_buf->buffer_length <= 0) {
		gdam_debug("wavetable %u: still loading...", buffer_id);
		return;
	}
	gtk_object_ref(GTK_OBJECT(cache_buf));
	if (wavetable->cache)
		gtk_object_unref(GTK_OBJECT(wavetable->cache));
	wavetable->cache = cache_buf;
	{
		gint64 buf_len_32 = (gint64)cache_buf->buffer_length << 32;
		if (wavetable->index_32 > buf_len_32)
			wavetable->index_32 %= buf_len_32;
	}
}
static void gdam_wavetable_arg_get_buffer_id(GtkObject*        object,
                                             GdamArg*          arg)
{
	GdamWavetable* wavetable = GDAM_WAVETABLE(object);
	if (wavetable->cache)
		GDAM_ARG_INT(arg) = wavetable->cache->buffer_id;
}
static void gdam_wavetable_arg_set_interpolating(GtkObject*        object,
                                                 GdamArg*          arg)
{
	GdamWavetable* wavetable = GDAM_WAVETABLE(object);
	wavetable->interpolating = GDAM_ARG_INT(arg);
}
static void gdam_wavetable_arg_get_interpolating(GtkObject*        object,
                                                 GdamArg*          arg)
{
	GdamWavetable* wavetable = GDAM_WAVETABLE(object);
	GDAM_ARG_INT(arg) = wavetable->interpolating;
}

static void gdam_wavetable_arg_set_terminates(GtkObject*        object,
                                                 GdamArg*          arg)
{
	GdamWavetable* wavetable = GDAM_WAVETABLE(object);
	wavetable->terminates = GDAM_ARG_INT(arg) ? 1 : 0;
}
static void gdam_wavetable_arg_get_terminates(GtkObject*        object,
                                                 GdamArg*          arg)
{
	GdamWavetable* wavetable = GDAM_WAVETABLE(object);
	GDAM_ARG_INT(arg) = wavetable->terminates;
}
static void gdam_wavetable_arg_set_use_overall_position
                                                (GtkObject*        object,
                                                 GdamArg*          arg)
{
	GdamWavetable* wavetable = GDAM_WAVETABLE(object);
	wavetable->use_overall_position = GDAM_ARG_INT(arg) ? 1 : 0;
}
static void gdam_wavetable_arg_get_use_overall_position
                                                (GtkObject*        object,
                                                 GdamArg*          arg)
{
	GdamWavetable* wavetable = GDAM_WAVETABLE(object);
	GDAM_ARG_INT(arg) = wavetable->use_overall_position;
}

static void gdam_wavetable_class_init(GdamSourceClass* to_init)
{
	GtkObjectClass* object_class = (GtkObjectClass*)to_init;
	to_init->read = gdam_wavetable_read;
	to_init->miss_data = gdam_wavetable_miss_data;
	to_init->get_position = gdam_wavetable_get_position;
	to_init->seek = gdam_wavetable_seek;
	object_class->destroy = gdam_wavetable_destroy;

	gdam_arg_register(object_class,
	                  "hertz",
			  GDAM_ARG_TYPE_DOUBLE,
			  gdam_wavetable_arg_set_hertz,
			  gdam_wavetable_arg_get_hertz,
	                  0);
	gdam_arg_register(object_class,
	                  "velocity",
			  GDAM_ARG_TYPE_DOUBLE,
			  gdam_wavetable_arg_set_velocity,
			  gdam_wavetable_arg_get_velocity,
	                  GDAM_ARG_VITAL);
	gdam_arg_register(object_class,
	                  "use_overall_position",
			  GDAM_ARG_TYPE_INT,
			  gdam_wavetable_arg_set_use_overall_position,
			  gdam_wavetable_arg_get_use_overall_position,
	                  GDAM_ARG_VITAL);
	gdam_arg_register(object_class,
	                  "buffer_id",
			  GDAM_ARG_TYPE_INT,
			  gdam_wavetable_arg_set_buffer_id,
			  gdam_wavetable_arg_get_buffer_id,
	                  GDAM_ARG_VITAL);
	gdam_arg_register(object_class,
	                  "interpolating",
			  GDAM_ARG_TYPE_INT,
			  gdam_wavetable_arg_set_interpolating,
			  gdam_wavetable_arg_get_interpolating,
	                  GDAM_ARG_VITAL);
	gdam_arg_register(object_class,
	                  "terminates",
			  GDAM_ARG_TYPE_INT,
			  gdam_wavetable_arg_set_terminates,
			  gdam_wavetable_arg_get_terminates,
	                  GDAM_ARG_VITAL);
}

static void gdam_wavetable_init(GdamWavetable* to_init)
{
	to_init->index_32 = 0;
	to_init->velocity_32 = (1LL << 32);
	to_init->interpolating = 1;
}

GtkType gdam_wavetable_get_type ()
{
	static GtkType wavetable_type = 0;
	if (!wavetable_type) {
		static const GtkTypeInfo wavetable_info = {
			"GdamWavetable",
			sizeof(GdamWavetable),
			sizeof(GdamSourceClass),
			(GtkClassInitFunc) gdam_wavetable_class_init,
			(GtkObjectInitFunc) gdam_wavetable_init,
			/* reserved_1*/ NULL,
			/* reserved_2*/ NULL,
			(GtkClassInitFunc) NULL
		};
		parent_class = gtk_type_class(GDAM_TYPE_SOURCE);
		parent_source_class = (GdamSourceClass*)parent_class;
		wavetable_type = gtk_type_unique(GDAM_TYPE_SOURCE,
		                                  &wavetable_info);
	}
	return wavetable_type;
}

static inline GdamSample wavetable_index   (GdamSample* buffer,
                                            gint64      index_32,
				            gint        buflen)
{
	return buffer[index_32 >> 32];
}

/* TODO: try to write inline assembly for this... */
static inline GdamSample wavetable_lerp   (GdamSample* buffer,
                                           gint64      index_32,
				           gint        buflen)
{
	GdamSample out;
	int index = (int)(index_32>>32);
	guint32 fraction = (guint32)(index_32);
	const GdamSample* a = buffer + index;
	const GdamSample* b;
	if (index == buflen - 1)
		b = buffer;
	else
		b = a + 1;
	out.s[0] =(gint16) (     (((gint64)(~fraction) * a->s[0])
	                        + ((gint64) (fraction) * b->s[0])) >> 32    );
	out.s[1] =(gint16) (     (((gint64)(~fraction) * a->s[1])
	                        + ((gint64) (fraction) * b->s[1])) >> 32    );
	return out;
}


#define	MAX_WAVETABLE_AVERAGE		3
static int       gdam_wavetable_read           (GdamSource*      input,
	                                        GdamSample*      samples,
						int              max_read)
{
	GdamWavetable* wavetable = GDAM_WAVETABLE(input);
	int buffer_len = wavetable->cache->buffer_length;
	GdamSample* buffer;
	gint64 index_32 = wavetable->index_32;
	gint64 velocity_32 = wavetable->velocity_32;
	gint64 buflen_32 = (gint64)buffer_len << 32;
	gboolean terminates = wavetable->terminates;
	int rv = max_read;

	if (!wavetable->cache)
		return 0;

	buffer = wavetable->cache->buffer;
	g_assert(0 <= index_32);
	g_assert(index_32 < ((gint64)buffer_len << 32));
	if (velocity_32 < 0) {
		if (wavetable->interpolating) {
			while (max_read-- > 0) {
				*samples++ = wavetable_lerp(buffer,
						  index_32,
						  buffer_len);
				index_32 += velocity_32;
				if (index_32 < 0) {
					if (terminates)
						break;
					index_32 += buflen_32;
				}
			}
		} else {
			while (max_read-- > 0) {
				*samples++ = wavetable_index(buffer,
						             index_32,
						             buffer_len);
				index_32 += velocity_32;
				if (index_32 < 0) {
					if (terminates)
						break;
					index_32 += buflen_32;
				}
			}
		}
	} else if (velocity_32 > 0) {
		if (wavetable->interpolating) {
			while (max_read-- > 0) {
				*samples++ = wavetable_lerp(buffer,
						  index_32,
						  buffer_len);
				index_32 += velocity_32;
				if (index_32 >= buflen_32) {
					if (terminates)
						break;
					index_32 -= buflen_32;
				}
			}
		} else {
			while (max_read-- > 0) {
				*samples++ = wavetable_index(buffer,
						             index_32,
						             buffer_len);
				index_32 += velocity_32;
				if (index_32 >= buflen_32) {
					if (terminates)
						break;
					index_32 -= buflen_32;
				}
			}
		}
	} else {
		/* um, velocity == 0??? */
		gdam_sample_zero(samples, rv);
		max_read = 0;
	}
	wavetable->index_32 = index_32;
	if (max_read <= 0) {
		wavetable->position += rv;
		return rv;
	}
	wavetable->position += (rv - max_read);
	return rv - max_read;
}

static void      gdam_wavetable_miss_data      (GdamSource*      input,
	                                        int              samples_lost)
{
	GdamWavetable* wavetable = GDAM_WAVETABLE(input);
	gint64 new_index = wavetable->index_32;
	new_index += (gint64)samples_lost * wavetable->velocity_32;
	new_index %= (((gint64)wavetable->cache->buffer_length) << 32);
	wavetable->index_32 = new_index;
	wavetable->position += samples_lost;
}

static GdamTime    gdam_wavetable_get_position     (GdamSource*   source)
{
	GdamWavetable* wavetable = GDAM_WAVETABLE(source);
	if (wavetable->use_overall_position)
		return wavetable->position;
	if (wavetable->velocity_32 >= wavetable->index_32)
		return 0;
	return wavetable->index_32 / wavetable->velocity_32;
}

static gboolean    gdam_wavetable_seek             (GdamSource*   source,
                                                    GdamTime      time,
						    int           whence)
{
	GdamWavetable* wavetable = GDAM_WAVETABLE(source);
	int buffer_length;
	if (wavetable->use_overall_position) {
		if (whence == 0)
			wavetable->position = time;
		else
			wavetable->position += time;
		return TRUE;
	}
	if (wavetable->cache == NULL)
		return FALSE;
	buffer_length = wavetable->cache->buffer_length;
	if (buffer_length >= wavetable->velocity_32)
		return FALSE;

	switch (whence) {
		case 0: case 2: {
			wavetable->index_32 = time * wavetable->velocity_32;
			wavetable->index_32 %= ((gint64)buffer_length << 32);
			break;
		}
		case 1: {
			wavetable->index_32 += time * wavetable->velocity_32;
			wavetable->index_32 %= ((gint64)buffer_length << 32);
			break;
		}
		default: g_warning("wavetable_seek: bad whence"); break;
	}
	while (wavetable->index_32 < 0)
		wavetable->index_32 += ((gint64)buffer_length << 32);
	return TRUE;
}

static void      gdam_wavetable_destroy            (GtkObject*       object)
{
	GdamWavetable* wavetable = GDAM_WAVETABLE(object);
	if (wavetable->cache) {
		GtkObject* buf = GTK_OBJECT(wavetable->cache);
		wavetable->cache = NULL;
		gtk_object_unref(buf);
	}
	parent_class->destroy(object);
}


#if 0
void gdam_wavetable_set_velocity(GdamWavetable*   wavetable,
                                 gint64           velocity_32)
{
	wavetable->velocity_32 = velocity_32;
}
void gdam_wavetable_set_hertz   (GdamWavetable*   wavetable,
                                 gdouble          hertz)
{
	int buffer_len = wavetable->cache->buffer_length;
	/* Broken down to prevent overflow. */
	gint64 velo = (gint64)(hertz * buffer_len / 44100.0);
	wavetable->velocity_32 = velo;
}
#endif
