/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/

#include "gdamflange.h"
#include <gdam/gdamnamespace.h>

static int gdam_flange_read(GdamSource*    source,
                            GdamSample*    buffer,
			    int            max_read)
{
	GdamFlange* flange = GDAM_FLANGE(source);
	GdamBufferingFilter* buffering_filter = GDAM_BUFFERING_FILTER(source);

	GdamSample* modulation = NULL;
	int num_read;
	int i;

	num_read = gdam_filter_sub_read(GDAM_FILTER(source), buffer, max_read);

	/* Fill modulation buffer. */
	if (num_read > 0) {
		int mod_read = 0;
		modulation = alloca(sizeof(GdamSample) * num_read);
		if (flange->modulator != NULL) {
			mod_read = gdam_source_read(flange->modulator,
			                            modulation,
						    num_read);
		}
		if (mod_read < num_read)
			gdam_sample_zero(modulation + mod_read,
			                 num_read - mod_read);
	}

	for (i = 0; i < num_read; i++) {
		int channel;
		gdam_buffering_filter_append_sample(buffering_filter,
		                                    buffer[i]);
		for (channel = 0; channel < 2; channel++) {
			int mod;
			int sum = 0;
			int echo;
			GdamFixed factor;
			mod = modulation[i].s[channel] - GDAM_SAMPLE_MIN;
			factor = (GdamFixed) ( ((gint64)mod
			                 * flange->delta_echo_factor)
					       >> (GDAM_SAMPLE_LOG+1) );
			factor += flange->min_echo_factor;

			if (factor < 0) {
				g_error("i=%d, channel=%d, mod=%d, factor=%6.6g, modulation[i]={%d,%d}, flange: min/delta=%6.6g/%6.6g", i, channel, mod, GDAM_FIXED_TO_DOUBLE(factor), modulation[i].s[0], modulation[i].s[1], GDAM_FIXED_TO_DOUBLE(flange->min_echo_factor), GDAM_FIXED_TO_DOUBLE(flange->delta_echo_factor)); }

			for (echo = 0; echo < flange->num_echoes; echo++) {
				int delay;
				int value;
				int volume;
				delay = flange->echoes[echo].delay;
				delay = (gint64)factor * delay / GdamFixed_ONE;
				value = gdam_buffering_filter_sample_at
					(buffering_filter, delay).s[channel];
				volume = flange->echoes[echo].volume;
				sum += GDAM_FIXED_MUL(volume, value);
			}
			buffer[i].s[channel] = GDAM_SAMPLE_CLAMP(sum);
		}
	}
	return num_read;
}

static void do_ensure_space(GdamFlange* flange)
{
	int max_delay = 0;
	int i;
	int required;
	GdamFixed max_echo_factor;
	for (i = 0; i < flange->num_echoes; i++) {
		int delay = flange->echoes[i].delay;
		if (delay > max_delay)
			max_delay = delay;
	}
	max_echo_factor = flange->min_echo_factor + flange->delta_echo_factor;
	required = (int)(((gint64)max_delay * max_echo_factor) / GdamFixed_ONE);
	gdam_buffering_filter_ensure_space((GdamBufferingFilter*)flange,
	                                   required + 1);
}

static void gdam_flange_arg_set_min_max(GtkObject* object, GdamArg* arg)
{
	gdouble* array = GDAM_ARG_DOUBLE_ARRAY(arg).array;
	int length = GDAM_ARG_DOUBLE_ARRAY(arg).length;
	GdamFlange* flange = GDAM_FLANGE(object);
	if (length != 2 || array[0] < 0.0 || array[0] > array[1]) {
		gdam_debug("array to flange::min_max was malformed");
		return;
	}
	if (array[1] > 100.0) {
		gdam_debug("flange: scale too large");
		return;
	}
	flange->min_echo_factor = GDAM_FIXED_FROM_DOUBLE(array[0]);
	flange->delta_echo_factor = GDAM_FIXED_FROM_DOUBLE(array[1]-array[0]);
	do_ensure_space(flange);
}
static void gdam_flange_arg_get_min_max(GtkObject* object, GdamArg* arg)
{
	gdouble* array;
	int max_echo_factor;
	GdamFlange* flange;
	array = g_new(gdouble, 2);
	GDAM_ARG_DOUBLE_ARRAY(arg).array = array;
	GDAM_ARG_DOUBLE_ARRAY(arg).length = 2;
	flange = GDAM_FLANGE(object);
	array[0] = GDAM_FIXED_TO_DOUBLE(flange->min_echo_factor);
	max_echo_factor = flange->min_echo_factor + flange->delta_echo_factor;
	array[1] = GDAM_FIXED_TO_DOUBLE(max_echo_factor);
}
static void gdam_flange_arg_set_echoes(GtkObject* object, GdamArg* arg)
{
	int num_echoes = GDAM_ARG_INT_ARRAY(arg).length / 2;
	GdamEchoElement* elements = g_new(GdamEchoElement, num_echoes * 2);
	int* array = GDAM_ARG_INT_ARRAY(arg).array;
	int i;
	GdamFlange* flange = GDAM_FLANGE(object);
	for (i = 0; i < num_echoes; i++) {
		elements[i].volume = array[2*i+0];
		elements[i].delay = array[2*i+1];
	}
	if (flange->echoes != NULL)
		g_free(flange->echoes);
	flange->echoes = elements;
	flange->num_echoes = num_echoes;
	do_ensure_space(flange);
}
static void gdam_flange_arg_get_echoes(GtkObject* object, GdamArg* arg)
{
	GdamFlange* flange = GDAM_FLANGE(object);
	int num_echoes = flange->num_echoes;
	int* array = g_new(int, num_echoes*2);
	int i;
	for (i = 0; i < num_echoes; i++) {
		array[2*i+0] = flange->echoes[i].volume;
		array[2*i+1] = flange->echoes[i].delay;
	}
	GDAM_ARG_INT_ARRAY(arg).length = num_echoes * 2;
	GDAM_ARG_INT_ARRAY(arg).array = array;
}


static void gdam_flange_arg_set_modulator(GtkObject* object, GdamArg* arg)
{
	int source_id = GDAM_ARG_INT(arg);
	GdamSource* sub = gdam_namespace_local_source(source_id);
	gdam_flange_set_modulator(GDAM_FLANGE(object), sub);
}
static void gdam_flange_arg_get_modulator(GtkObject* object, GdamArg* arg)
{
	GdamSource* modulator = GDAM_FLANGE(object)->modulator;
	GDAM_ARG_INT(arg) = modulator ? modulator->source_id : 0;
}

static void gdam_flange_init(GdamFlange* flange)
{
	flange->min_echo_factor = GdamFixed_ONE;
}
static void gdam_flange_class_init(GdamFlangeClass* flange_class)
{
	GtkObjectClass* object_class;

	GDAM_SOURCE_CLASS(flange_class)->read = gdam_flange_read;

	object_class = GTK_OBJECT_CLASS(flange_class);
	gdam_arg_register(object_class,
	                  "min_max",
			  GDAM_ARG_TYPE_DOUBLE_ARRAY,
			  gdam_flange_arg_set_min_max,
			  gdam_flange_arg_get_min_max,
			  GDAM_ARG_VITAL);
	gdam_arg_register(object_class,
	                  "echoes",
			  GDAM_ARG_TYPE_INT_ARRAY,
			  gdam_flange_arg_set_echoes,
			  gdam_flange_arg_get_echoes,
			  GDAM_ARG_VITAL);
	gdam_arg_register(object_class,
	                  "modulator",
			  GDAM_ARG_TYPE_INT,
			  gdam_flange_arg_set_modulator,
			  gdam_flange_arg_get_modulator,
			  GDAM_ARG_VITAL);
}

GtkType gdam_flange_get_type()
{
	static GtkType flange_type = 0;
	if (!flange_type) {
		static const GtkTypeInfo flange_info = {
			"GdamFlange",
			sizeof(GdamFlange),
			sizeof(GdamFlangeClass),
			(GtkClassInitFunc) gdam_flange_class_init,
			(GtkObjectInitFunc) gdam_flange_init,
			/* reserved_1 */ NULL,
			/* reserved_2 */ NULL,
			(GtkClassInitFunc) NULL
		};
		GtkType parent = GDAM_TYPE_BUFFERING_FILTER;
		flange_type = gtk_type_unique(parent,
		                                 &flange_info);
	}
	return flange_type;
}
void gdam_flange_set_modulator(GdamFlange* flange, GdamSource* mod)
{
	if (flange->modulator != NULL) {
		GdamSource* old_mod = flange->modulator;
		flange->modulator = NULL;
		old_mod->parent = NULL;
		gtk_object_unref(GTK_OBJECT(old_mod));
	}
	if (mod != NULL) {
		GtkObject* object = GTK_OBJECT(mod);
		gtk_object_ref(object);
		gtk_object_sink(object);
	}
	flange->modulator = mod;
}
