/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#include <gtk/gtk.h>
#include <glade/glade.h>
#include <stdlib.h>
//#include "gdamlauncher.h"
#include "gdamskin.h"
#include "gdamcrossfadeskin.h"
#include "gdamdeviceswitcherskin.h"
#include "gdampacker.h"
#include "gdamdefault.h"
#include "gdamdnd.h"
#include <gdam/component/gdamutil.h>
#include <gdam/component/gdamsongdatabase.h>
#include <gdam/component/gdambuffermanager.h>

char* gdam_rcpath = 0;

static const char* default_rcfiles[] = { 
                                  	"~/.gdam/gdamrc", 
					GDAM_SKIN_PREFIX "/default/default.rc",
					".gdamrc",
					"~/.gdamrc",
				 	"/usr/local/lib/gdam/skins/default.rc",
				  	"/etc/gdam/gdamrc", 
					"default.rc",
				  	"~/.gdam/default.rc", 
					0 };

static void usage(const char* progname)
{
	fprintf(stderr, "usage: %s [--device device_to_open]\n"
	                "\n"
			"Options:\n"
			"\t--device device   : set the device\n"
			"\t--rcfile rcfile   : use alternate rc file\n"
			"\t--rcpath rcpath   : use alternate path when "
						"searching for rc file\n"
			"\t--database db     : specify a path to the song "
						"database [abbrev: -d]\b"
			"\nUndocumented: drag2 drag3\n",
			progname
			);
	exit(1);
}

static void open_device(gpointer list_data, gpointer user_data)
{
	gdam_toplevel_open_device(GDAM_TOPLEVEL(user_data),
	                          (const char*)list_data,
				  NULL,
				  NULL,
				  NULL);
}

#if 0
static gboolean is_doc_from_glade(xmlDoc* doc)
{
	return TRUE;
}
#endif


int main(int argc, char** argv)
{
	//GdamLauncher* launcher;
	GdamToplevel* toplevel;
	GdamChannel* channel;
	GdamSongDatabase* db;
	int i;
	int skins_loaded = 0;
	GSList* devices = NULL;
	const char* rcfile = NULL;
	static GSList* xml_files = NULL;

	/* XXX: instead of this block, we should break apart the
	 *      argument parsing/connections api of gdam_channel_init
	 *      and do all of the parsing prior to connecting.
	 *
	 *      But for now...
	 */
	if (argv[1] != NULL) {
		if (strcmp(argv[1], "--help") == 0 
		 || strcmp(argv[1], "-h")     == 0
		 || strcmp(argv[1], "-help")  == 0) {
			usage(argv[0]);
			return 0;
		}
	}

	if (!gdam_channel_init(&argc, &argv, &channel)) {
		g_error("error creating connection to client");
		return 1;
	}
	if (argv[1] != NULL) {
		if (strcmp(argv[1], "-v") == 0
		 || strcmp(argv[1], "--version") == 0) {
		 	printf(
			     "gdam-launcher version %s.\n"
		             "(C) Copyright Dave Benson, Geoff Matters. 2000.\n"
			     "\n"
			     "THIS PROGRAM COMES WITH NO WARRANTY! "
			     "USE AT YOUR OWN RISK.\n",
			     VERSION);
			exit(0);
		}
	}
	g_return_val_if_fail(channel != 0, 1);
	gtk_init(&argc, &argv);
        glade_init();

	for (i = 1; i < argc; i++) {
		char* opt = argv[i];
		if (argv[i][0] != '-') {
			xml_files = g_slist_prepend(xml_files, argv[i]);
			continue;
		}
		while (*opt == '-')
			opt++;
		if (strcmp(opt, "h") == 0 || strcmp(opt, "help") == 0)
			usage(argv[0]);
		if (strcmp(opt, "device") == 0) {
			char* device_name = argv[++i];
			if (device_name == NULL) {
				g_warning("--device needs arguments");
				return 1;
			}
			devices = g_slist_append(devices, device_name);
			continue;
		}
		if (strcmp(opt, "rcfile") == 0) {
			rcfile = argv[++i];
			continue;
		}
		if (strcmp(opt, "rcpath") == 0) {
			gdam_rcpath = argv[++i];
			continue;
		}
		if (strcmp(opt, "drag2") == 0) {
			gdam_drag_mask = GDK_BUTTON2_MASK;
			continue;
		}
		if (strcmp(opt, "drag3") == 0) {
			gdam_drag_mask = GDK_BUTTON3_MASK;
			continue;
		}
		if (strcmp(opt, "drag23") == 0) {
			gdam_drag_mask = GDK_BUTTON2_MASK|GDK_BUTTON3_MASK;
			continue;
		}
		if (strcmp(opt, "d") == 0 || strcmp(opt, "database") == 0) {
			++i;
			if (argv[i] == NULL) {
				g_error("%s must by followed by an argument",
					argv[i-1]);
			}
			gdam_song_database_default_filename = argv[i];
			continue;
		}
		g_warning("unknown option: %s", argv[i]);
		usage(argv[0]);
	}
	if (!rcfile && !(rcfile=gdam_default_file_find(default_rcfiles))) {
		fprintf(stderr, "Warning: no rcfile found.  "
		                "Colors will be gray.\n");
	} else if (!gdam_file_exists(rcfile)){
		fprintf(stderr, "Warning: rcfile (%s) not found.\n", rcfile);
	} else {
		fprintf(stderr, "Using rcfile %s.\n", rcfile);
		gtk_rc_parse(rcfile);
	}

	if (devices == NULL) {
		char* device_name = getenv("GDAM_DEVICE");
		if (device_name)
			devices = g_slist_append(devices, device_name);
	}

	if (devices == NULL) {
		g_message("Using /dev/dsp since no --device was given");
		devices = g_slist_append(devices, "/dev/dsp");
	}

	(void) gtk_type_class(GDAM_TYPE_PACKER);
	(void) gtk_type_class(GDAM_TYPE_CROSSFADE_SKIN);
	(void) gtk_type_class(GDAM_TYPE_DEVICE_SWITCHER_SKIN);

	/* Get triangle, square, sine waves. */
	gdam_buffer_manager_init(channel);

	toplevel = gdam_toplevel_get_global(channel);
	g_slist_foreach(devices, open_device, toplevel);

	if (xml_files == NULL)
		xml_files = g_slist_prepend(NULL, "launcher.xml");

	for (     ; xml_files != NULL; xml_files = xml_files->next) {
		const char* xml_file = xml_files->data;
		const char* fname;
		fname = gdam_skin_search_for_file(xml_file);
		if (fname == NULL) {
			g_warning("couldn't find xml file %s", fname);
			continue;
		}
		if (gdam_skin_open(fname, channel))
			skins_loaded++;
	}

	if (skins_loaded == 0)
		g_error("no skins were successfully made:  aborting");
	db = gdam_song_database_system();
	if (g_hash_table_size(db->name_lookup) == 0){
		const char* fname;
		fname = gdam_skin_search_for_file("songdbwarning.xml");
		if (fname == NULL) {
			g_warning("couldn't find xml file %s", fname);
		}
		else
			gdam_skin_open(fname, channel);
	}
	gtk_main();
	return 0;
}
